#include <Adafruit_NeoPixel.h>
#include "WiFi.h"

// NTP settings
#define TIMEZONE 9 // timezone (GMT = 0, Japan = 9)
#define NTP_SERVER "pool.ntp.org"

#define WIFI_SMARTCONFIG true

#if !WIFI_SMARTCONFIG
// if you do not use smartConfifg, please specify SSID and password here
#define WIFI_SSID "SSID" // your WiFi's SSID
#define WIFI_PASS "PASS" // your WiFi's password
#endif

#define LED_PIN 2
Adafruit_NeoPixel pixels(1, LED_PIN, NEO_GRB + NEO_KHZ800);

#define STEPS_PER_ROTATION 4096 // steps for a full turn of minute rotor
#define POS 14
#define MOTOR_NUM 3
#define MOTOR_PIN 4
#define MIN1 0
#define MIN10 1
#define HOUR 2

// minimum wait for a single step of stepper
#define DELAY 1

// ports used to control the stepper motor
// if your motor rotate to the opposite direction, 
// reverse the order of the port.
int port[MOTOR_NUM][MOTOR_PIN] = {
  {7, 6, 5, 4},   // 1min digit
  {10, 8, 1, 0},  // 10min digit
  {21, 19, 18, 9} // hour digit
};

// adjust initial position slightly
int initPos[MOTOR_NUM] = {50, 70, 70};

long stepToGo[MOTOR_NUM] = {0};

// sequence of stepper motor control
int seq[8][MOTOR_PIN] = {
  {  LOW, HIGH, HIGH,  LOW},
  {  LOW,  LOW, HIGH,  LOW},
  {  LOW,  LOW, HIGH, HIGH},
  {  LOW,  LOW,  LOW, HIGH},
  { HIGH,  LOW,  LOW, HIGH},
  { HIGH,  LOW,  LOW,  LOW},
  { HIGH, HIGH,  LOW,  LOW},
  {  LOW, HIGH,  LOW,  LOW}
};

void rotateSet(int num, long step) {
  stepToGo[num] += step;
}

void rotateGo(void) {
  static long phase[MOTOR_NUM] = {0};
  int dt = 20;
  int done = 0;
  int delta;

  while(!done) {
    done = 1;
    for(int m = 0; m < MOTOR_NUM; m++) {
      long step = stepToGo[m];
      if(step == 0) {
        for(int i = 0; i < MOTOR_PIN; i++) {
          digitalWrite(port[m][i], LOW);
        }
        continue;
      }
      else if(step > 0) {
        stepToGo[m]--;
        delta = 1;
      }
      else {
        stepToGo[m]++;
        delta = 7;
      }
      done = 0;
      phase[m] = (phase[m] + delta) % 8;
      for(int i = 0; i < MOTOR_PIN; i++) {
        digitalWrite(port[m][i], seq[phase[m]][i]);
      }
    }
    delay(dt);
    if(dt > DELAY) {
      dt--; // accelerate
    }
  }
  // power cut
  for(int m = 0; m < MOTOR_NUM; m++) {
    for(int i = 0; i < MOTOR_PIN; i++) {
      digitalWrite(port[m][i], LOW);
    }
  }
}

// order of playing cards
int order[MOTOR_NUM][POS] = {
  {0, 10, 5, 3, 8, 6, 2, 13, 9, 1, 11, 4, 12, 7}, // 1min digit
  {0, 13, 7, 10, 8, 12, 6, 3, 11, 5, 1, 4, 2, 9}, // 10min digit
  {0, 2, 7, 10, 4, 12, 5, 8, 3, 11, 1, 6, 9, 13}  // hour digit
};

int currentPos[MOTOR_NUM] = {0};

void setNum(int m, int n) {
  int i, pos, newPos, advance;

  for(i = 0; i < POS; i++) {
    if(order[m][i] == n)
      newPos = i;
  }
  pos = (newPos * STEPS_PER_ROTATION) / POS;
  advance = (pos - currentPos[m] + STEPS_PER_ROTATION) % STEPS_PER_ROTATION; // always CW rotation
  rotateSet(m, advance);
  currentPos[m] = pos;
}

void setup() {
  Serial.begin(115200);
  Serial.println("start");
  
  for(int m = 0; m < MOTOR_NUM; m++) {
    for(int i = 0; i < MOTOR_PIN; i++) {
      pinMode(port[m][i], OUTPUT);
    }
  }
  pixels.setPixelColor(0, pixels.Color(0, 255, 0)); // green
  pixels.show();
  for(int m = 0; m < MOTOR_NUM; m++) {
    rotateSet(m, -STEPS_PER_ROTATION); // reverse rotation to endstop
  }
  rotateGo();
  for(int m = 0; m < MOTOR_NUM; m++) {
    rotateSet(m, STEPS_PER_ROTATION / 2 + initPos[m]); // show 0
  }
  rotateGo();
  
  pixels.setPixelColor(0, pixels.Color(0, 0, 255)); // blue
  pixels.show();

  wifiSetup();

  getNTP();

  pixels.setPixelColor(0, pixels.Color(0, 0, 0)); // turn off
  pixels.show();
}

void loop() {
  static int prevMin = -1;
  static int prevDay = -1;
  int hour;
  struct tm tmtime;
    
  getLocalTime(&tmtime);
  if(prevDay > 0 && tmtime.tm_mday != prevDay) {
    // ESP.restart(); // reset itself at 0:00 to get accurate time
    getNTP();
  }
  prevDay = tmtime.tm_mday;

  if(prevMin == tmtime.tm_min) {
    return;
  }
  prevMin = tmtime.tm_min;
  
  hour = tmtime.tm_hour % 12;
  if(hour == 0) hour = 12;
  
  setNum(HOUR, hour);
  setNum(MIN10, tmtime.tm_min / 10);
  setNum(MIN1, tmtime.tm_min % 10);
  rotateGo();
  delay(900);
  
}

//////////////////////////////
// WiFi and NTP section
//////////////////////////////

void getNTP(void) {
  for(int i = 0; WiFi.status() != WL_CONNECTED; i++) {
    if(i > 30) {
      ESP.restart();
    }
    Serial.println("Waiting for WiFi connection..");
    delay(1000);
  }

  configTime(TIMEZONE * 3600L, 0, NTP_SERVER);
  printLocalTime();
}

void printLocalTime() {
  struct tm timeinfo;
  if (!getLocalTime(&timeinfo)) {
    Serial.println("Failed to obtain time");
    return;
  }
  Serial.println(&timeinfo, "%A, %Y-%m-%d %H:%M:%S");
}


void wifiSetup() {
  WiFi.mode(WIFI_STA);
#if WIFI_SMARTCONFIG
  WiFi.begin();
#else
  WiFi.begin(WIFI_SSID, WIFI_PASS);
#endif

  for (int i = 0; ; i++) {
    Serial.println("Connecting to WiFi...");
    delay(1000);
    if (WiFi.status() == WL_CONNECTED) {
      break;
    }
#if WIFI_SMARTCONFIG
  if(i > 30)
    break;
#endif    
  }

#if WIFI_SMARTCONFIG
  if (WiFi.status() != WL_CONNECTED) {
    WiFi.mode(WIFI_AP_STA);
    WiFi.beginSmartConfig();

    //Wait for SmartConfig packet from mobile
    Serial.println("Waiting for SmartConfig.");
    while (!WiFi.smartConfigDone()) {
      Serial.print(".");
      pixels.setPixelColor(0, pixels.Color(0, 255, 0)); // red
      pixels.show();
      delay(1000);
    }

    Serial.println("");
    Serial.println("SmartConfig received.");

    //Wait for WiFi to connect to AP
    Serial.println("Waiting for WiFi");
    while (WiFi.status() != WL_CONNECTED) {
      delay(1000);
      Serial.print(",");
    }
  }
  Serial.println("WiFi Connected.");
#endif

  Serial.print("SSID: ");
  Serial.println(WiFi.SSID());
  Serial.print("IP Address: ");
  Serial.println(WiFi.localIP());
}
